import numpy as np
import pytest
import resqpy.olio.write_data as wd
from pytest_mock import MockerFixture


def test_write_pure_binary_data(mocker: MockerFixture, caplog):
    # Arrange
    test_array = np.array([[[0, 0], [0, 0]], [[0, 0], [0, 0]]])
    open_mock = mocker.mock_open()
    fileno_mock = mocker.Mock(return_value = 1, name = 'fileno_mock')
    open_mock.return_value.fileno = fileno_mock

    mocker.patch("builtins.open", open_mock)
    expected_calls = [mocker.call('test', 'wb'), mocker.call().__enter__()]

    # Act
    wd.write_pure_binary_data('test', test_array)

    # Assert
    open_mock.assert_has_calls(expected_calls)
    assert 'Binary data file test created' in caplog.text


def test_write_array_to_ascii_file_decimals_equal_0(mocker: MockerFixture):
    # Arrange
    test_extent = np.array([2, 2, 2])
    test_array = np.array([[[0, 0], [0, 0]], [[0, 0], [0, 0]]])
    open_mock = mocker.mock_open()
    mocker.patch("builtins.open", open_mock)
    expected_calls = [
        mocker.call('test', 'w'),
        mocker.call().__enter__(),
        mocker.call().write('! Data written by write_array_to_ascii_file() python function\n'),
        mocker.call().write('! Extent of array is: [2, 2, 2]\n'),
        mocker.call().write('! Maximum 20 data items per line\n'),
        mocker.call().write('\n'),
        mocker.call().write('0'),
        mocker.call().write('\t'),
        mocker.call().write('0'),
        mocker.call().write('\n'),
        mocker.call().write('\n'),
        mocker.call().write('0'),
        mocker.call().write('\t'),
        mocker.call().write('0'),
        mocker.call().write('\n'),
        mocker.call().write('\n'),
        mocker.call().write('0'),
        mocker.call().write('\t'),
        mocker.call().write('0'),
        mocker.call().write('\n'),
        mocker.call().write('\n'),
        mocker.call().write('0'),
        mocker.call().write('\t'),
        mocker.call().write('0'),
        mocker.call().write('\n'),
        mocker.call().__exit__(None, None, None)
    ]

    # Act
    wd.write_array_to_ascii_file('test', test_extent, test_array, decimals = 0)

    # Assert
    open_mock.assert_has_calls(expected_calls)


def test_write_array_to_ascii_file_binary(mocker: MockerFixture, caplog):
    # Arrange
    test_extent = np.array([2, 2, 2])
    test_array = np.array([[[0, 0], [0, 0]], [[0, 0], [0, 0]]])
    open_mock = mocker.mock_open()
    fileno_mock = mocker.Mock(return_value = 1, name = 'fileno_mock')
    open_mock.return_value.fileno = fileno_mock

    mocker.patch("builtins.open", open_mock)
    expected_calls = [mocker.call('test.db', 'wb'), mocker.call().__enter__()]

    # Act
    wd.write_array_to_ascii_file('test', test_extent, test_array, use_binary = True, binary_only = True)

    # Assert
    open_mock.assert_has_calls(expected_calls)
    assert 'Failed to write data to binary file test' not in caplog.text


def test_write_array_to_ascii_file_no_headers(mocker: MockerFixture):
    # Arrange
    test_extent = np.array([2, 2, 2])
    test_array = np.array([[[0, 0], [0, 0]], [[0, 0], [0, 0]]])
    open_mock = mocker.mock_open()
    mocker.patch("builtins.open", open_mock)
    unexpected_calls = [
        mocker.call().write('! Data written by write_array_to_ascii_file() python function\n'),
        mocker.call().write('! Extent of array is: [2, 2, 2]\n'),
        mocker.call().write('! Maximum 20 data items per line\n')
    ]

    # Act
    wd.write_array_to_ascii_file('test', test_extent, test_array, headers = False)

    # Assert
    assert unexpected_calls not in open_mock.mock_calls


def test_write_array_to_ascii_file_bool(mocker: MockerFixture):
    # Arrange
    test_extent = np.array([2, 2, 2])
    test_array = np.array([[[0, 1], [0, 0]], [[1, 0], [1, 0]]])
    open_mock = mocker.mock_open()
    mocker.patch("builtins.open", open_mock)
    expected_calls = [
        mocker.call('test', 'w'),
        mocker.call().__enter__(),
        mocker.call().write('! Data written by write_array_to_ascii_file() python function\n'),
        mocker.call().write('! Extent of array is: [2, 2, 2]\n'),
        mocker.call().write('! Maximum 20 data items per line\n'),
        mocker.call().write('\n'),
        mocker.call().write('0'),
        mocker.call().write('\t'),
        mocker.call().write('1'),
        mocker.call().write('\n'),
        mocker.call().write('\n'),
        mocker.call().write('0'),
        mocker.call().write('\t'),
        mocker.call().write('0'),
        mocker.call().write('\n'),
        mocker.call().write('\n'),
        mocker.call().write('1'),
        mocker.call().write('\t'),
        mocker.call().write('0'),
        mocker.call().write('\n'),
        mocker.call().write('\n'),
        mocker.call().write('1'),
        mocker.call().write('\t'),
        mocker.call().write('0'),
        mocker.call().write('\n'),
        mocker.call().__exit__(None, None, None)
    ]

    # Act
    wd.write_array_to_ascii_file('test', test_extent, test_array, data_type = 'bool')

    # Assert
    open_mock.assert_has_calls(expected_calls)


def test_write_array_to_ascii_file_append(mocker: MockerFixture):
    # Arrange
    test_extent = np.array([2, 2, 2])
    test_array = np.array([[[0, 0], [0, 0]], [[0, 0], [0, 0]]])
    open_mock = mocker.mock_open()
    mocker.patch("builtins.open", open_mock)
    expected_calls = [mocker.call('test', 'a')]

    # Act
    wd.write_array_to_ascii_file('test', test_extent, test_array, append = True)

    # Assert
    open_mock.assert_has_calls(expected_calls)


def test_write_array_to_ascii_file_space_separated(mocker: MockerFixture):
    # Arrange
    test_extent = np.array([2, 2, 2])
    test_array = np.array([[[0, 0], [0, 0]], [[0, 0], [0, 0]]])
    open_mock = mocker.mock_open()
    mocker.patch("builtins.open", open_mock)
    expected_calls = [
        mocker.call('test', 'w'),
        mocker.call().__enter__(),
        mocker.call().write('! Data written by write_array_to_ascii_file() python function\n'),
        mocker.call().write('! Extent of array is: [2, 2, 2]\n'),
        mocker.call().write('! Maximum 20 data items per line\n'),
        mocker.call().write('\n'),
        mocker.call().write('0.000'),
        mocker.call().write(' '),
        mocker.call().write('0.000'),
        mocker.call().write('\n'),
        mocker.call().write('\n'),
        mocker.call().write('0.000'),
        mocker.call().write(' '),
        mocker.call().write('0.000'),
        mocker.call().write('\n'),
        mocker.call().write('\n'),
        mocker.call().write('0.000'),
        mocker.call().write(' '),
        mocker.call().write('0.000'),
        mocker.call().write('\n'),
        mocker.call().write('\n'),
        mocker.call().write('0.000'),
        mocker.call().write(' '),
        mocker.call().write('0.000'),
        mocker.call().write('\n'),
        mocker.call().__exit__(None, None, None)
    ]

    # Act
    wd.write_array_to_ascii_file('test', test_extent, test_array, space_separated = True)

    # Assert
    open_mock.assert_has_calls(expected_calls)


def test_write_array_to_ascii_file_nan_substitute(mocker: MockerFixture):
    # Arrange
    test_extent = np.array([2, 2, 2])
    test_array = np.array([[[1, np.nan], [3, 2]], [[1, 5], [np.nan, np.nan]]])
    open_mock = mocker.mock_open()
    mocker.patch("builtins.open", open_mock)
    expected_calls = [
        mocker.call('test', 'w'),
        mocker.call().__enter__(),
        mocker.call().write('! Data written by write_array_to_ascii_file() python function\n'),
        mocker.call().write('! Extent of array is: [2, 2, 2]\n'),
        mocker.call().write('! Maximum 20 data items per line\n'),
        mocker.call().write('\n'),
        mocker.call().write('1.000'),
        mocker.call().write('\t'),
        mocker.call().write('0.000'),
        mocker.call().write('\n'),
        mocker.call().write('\n'),
        mocker.call().write('3.000'),
        mocker.call().write('\t'),
        mocker.call().write('2.000'),
        mocker.call().write('\n'),
        mocker.call().write('\n'),
        mocker.call().write('1.000'),
        mocker.call().write('\t'),
        mocker.call().write('5.000'),
        mocker.call().write('\n'),
        mocker.call().write('\n'),
        mocker.call().write('0.000'),
        mocker.call().write('\t'),
        mocker.call().write('0.000'),
        mocker.call().write('\n'),
        mocker.call().__exit__(None, None, None)
    ]

    # Act
    wd.write_array_to_ascii_file('test', test_extent, test_array, nan_substitute_value = 0)

    # Assert
    open_mock.assert_has_calls(expected_calls)


def test_write_array_to_ascii_file_blank_line_after_j_block(mocker: MockerFixture):
    # Arrange
    test_extent = np.array([2, 2, 2])
    test_array = np.array([[[0, 0], [0, 0]], [[0, 0], [0, 0]]])
    open_mock = mocker.mock_open()
    mocker.patch("builtins.open", open_mock)
    expected_calls = [
        mocker.call('test', 'w'),
        mocker.call().__enter__(),
        mocker.call().write('! Data written by write_array_to_ascii_file() python function\n'),
        mocker.call().write('! Extent of array is: [2, 2, 2]\n'),
        mocker.call().write('! Maximum 20 data items per line\n'),
        mocker.call().write('\n'),
        mocker.call().write('0.000'),
        mocker.call().write('\t'),
        mocker.call().write('0.000'),
        mocker.call().write('\n'),
        mocker.call().write('\n'),
        mocker.call().write('0.000'),
        mocker.call().write('\t'),
        mocker.call().write('0.000'),
        mocker.call().write('\n'),
        mocker.call().write('\n'),
        mocker.call().write('\n'),
        mocker.call().write('0.000'),
        mocker.call().write('\t'),
        mocker.call().write('0.000'),
        mocker.call().write('\n'),
        mocker.call().write('\n'),
        mocker.call().write('0.000'),
        mocker.call().write('\t'),
        mocker.call().write('0.000'),
        mocker.call().write('\n'),
        mocker.call().write('\n'),
        mocker.call().__exit__(None, None, None)
    ]

    # Act
    wd.write_array_to_ascii_file('test', test_extent, test_array, blank_line_after_j_block = True)

    # Assert
    open_mock.assert_has_calls(expected_calls)



import pandas as pd
import pytest
import resqpy.olio.write_faults as wf
from pytest_mock import MockerFixture


@pytest.mark.parametrize("test_df", [pd.DataFrame(), None])
def test_no_data_in_faults(tmp_path, test_df):
    # Act & Assert
    with pytest.raises(AssertionError):
        wf.write_faults_nexus(f'{tmp_path}/test.txt', test_df)


def test_writing_faults_to_file(tmp_path, mocker: MockerFixture):
    # Arrange
    test_df = pd.DataFrame({
        'name': ['fault1', 'fault2'],
        'i1': [720, 305],
        'i2': [875, 342],
        'j1': [311, 32],
        'j2': [103, 800],
        'k1': [791, 847],
        'k2': [994, 494],
        'face': ['I+', 'I+']
    })
    open_mock = mocker.mock_open()
    mocker.patch("builtins.open", open_mock)
    expected_calls = [
        mocker.call(f'{tmp_path}/test.txt', 'w'),
        mocker.call().__enter__(),
        mocker.call().write('\nMULT\tTX\tALL\tPLUS\tMULT\n'),
        mocker.call().write('\tGRID\tROOT\n'),
        mocker.call().write('\tFNAME\tfault1\n'),
        mocker.call().write('\t720\t875\t311\t103\t791\t994\t1.0\n'),
        mocker.call().write('\nMULT\tTX\tALL\tPLUS\tMULT\n'),
        mocker.call().write('\tGRID\tROOT\n'),
        mocker.call().write('\tFNAME\tfault2\n'),
        mocker.call().write('\t305\t342\t32\t800\t847\t494\t1.0\n'),
        mocker.call().__exit__(None, None, None)
    ]

    # Act
    wf.write_faults_nexus(f'{tmp_path}/test.txt', test_df)

    # Assert
    open_mock.assert_has_calls(expected_calls)


def test_log_info(tmp_path, caplog):
    # Arrange
    test_df = pd.DataFrame({
        'name': ['fault1', 'fault2'],
        'i1': [720, 305],
        'i2': [875, 342],
        'j1': [311, 32],
        'j2': [103, 800],
        'k1': [791, 847],
        'k2': [994, 494],
        'face': ['I+', 'I+']
    })

    # Act
    wf.write_faults_nexus(f'{tmp_path}/test.txt', test_df)

    # Assert
    for record in caplog.records:
        assert record.levelname == "INFO"
    assert f'writing FNAME data in Nexus format to file: {tmp_path}/test.txt in caplog.text'


def test_fault_name_greater_than_256_characters(tmp_path, caplog):
    # Arrange
    long_name = 'f' * 257
    test_df = pd.DataFrame({
        'name': ['fault1', long_name],
        'i1': [720, 305],
        'i2': [875, 342],
        'j1': [311, 32],
        'j2': [103, 800],
        'k1': [791, 847],
        'k2': [994, 494],
        'face': ['I+', 'I+']
    })

    # Act
    wf.write_faults_nexus(f'{tmp_path}/test.txt', test_df)

    # Assert
    assert f'exported fault name longer than Nexus limit of 256 characters: {long_name}' \
           in caplog.text



# test module for the derived RelPerm class located in resqpy.olio.dataframe.py

import os

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

import resqpy.model as rq
from resqpy.olio.relperm import (RelPerm, relperm_parts_in_model, text_to_relperm_dict)


def test_col_headers(tmp_path):
    epc = os.path.join(tmp_path, 'model.epc')
    model = rq.new_model(epc)
    np_df = np.array([[0.0, 0.0, 1.0, 0], [0.04, 0.015, 0.87, np.nan], [0.12, 0.065, 0.683, np.nan],
                      [0.25, 0.205, 0.35, np.nan], [0.45, 0.55, 0.018, np.nan], [0.74, 1.0, 0.0, 1e-06]])
    df_cols1 = ['Sw', 'Krg', 'Kro', 'Pc']
    df_cols2 = ['Sg', 'Krg', 'Krw', 'Pc']
    df_cols3 = ['Sg', 'Krg', 'Pc', 'Kro']
    df_cols4 = ['Sg', 'krg', 'Kro', 'pC']
    df1 = pd.DataFrame(np_df, columns = df_cols1)
    df2 = pd.DataFrame(np_df, columns = df_cols2)
    df3 = pd.DataFrame(np_df, columns = df_cols3)
    df4 = pd.DataFrame(np_df, columns = df_cols4)
    phase_combo1 = 'gas-oil'
    phase_combo2 = None

    with pytest.raises(ValueError) as excval1:
        RelPerm(model = model, df = df1, phase_combo = phase_combo1)
    assert "incorrect saturation column name and/or multiple saturation columns exist" in str(excval1.value)

    with pytest.raises(ValueError) as excval2:
        RelPerm(model = model, df = df2, phase_combo = phase_combo1)
    assert "incorrect column name(s) {'Krw'}" in str(excval2.value)

    with pytest.raises(ValueError) as excval3:
        RelPerm(model = model, df = df3, phase_combo = phase_combo1)
    assert "capillary pressure data should be in the last column of the dataframe" in str(excval3.value)

    relperm_obj = RelPerm(model = model, df = df4, phase_combo = phase_combo2)
    assert relperm_obj.phase_combo == 'gas-oil'


def test_missing_vals(tmp_path):
    epc = os.path.join(tmp_path, 'model.epc')
    model = rq.new_model(epc)
    np_df = np.array([[0.0, 0.0, 1.0, 0], [0.04, 0.015, 0.87, np.nan], [0.12, np.nan, 0.683, np.nan],
                      [0.25, 0.205, 0.35, np.nan], [0.45, 0.55, 0.018, np.nan], [0.74, 1.0, np.nan, 1e-06]])
    df_cols = ['Sg', 'Krg', 'Kro', 'Pc']
    df = pd.DataFrame(np_df, columns = df_cols)
    phase_combo = 'gas-oil'

    with pytest.raises(Exception) as excval:
        RelPerm(model = model, df = df, phase_combo = phase_combo)
    assert "missing values found in Krg column" in str(excval.value)


def test_monotonicity(tmp_path):
    epc = os.path.join(tmp_path, 'model.epc')
    model = rq.new_model(epc)
    np_df = np.array([[0.0, 0.0, 1.0, 0], [0.04, 0.015, 0.87, np.nan], [0.12, 0.065, 0.683, np.nan],
                      [0.25, 0.205, 0.35, np.nan], [0.85, 0.55, 0.018, 1e-06], [0.74, 1.0, 0.0, 1e-07]])
    df_cols = ['Sg', 'Krg', 'Kro', 'Pc']
    df = pd.DataFrame(np_df, columns = df_cols)
    phase_combo = 'gas-oil'

    with pytest.raises(ValueError) as excval:
        RelPerm(model = model, df = df, phase_combo = phase_combo)
    assert "('Sg', 'Krg', 'Kro') combo is not monotonic" in str(excval.value)

    df['Sg'] = [0.0, 0.04, 0.12, 0.25, 0.45, 0.74]
    with pytest.raises(ValueError) as excval1:
        RelPerm(model = model, df = df, phase_combo = phase_combo)
    assert "Pc values are not monotonic" in str(excval1.value)


def test_range(tmp_path):
    epc = os.path.join(tmp_path, 'model.epc')
    model = rq.new_model(epc)
    np_df = np.array([[0.0, 0.0, 1.0, 0], [0.04, 0.015, 0.87, np.nan], [0.12, 0.065, 0.683, np.nan],
                      [0.25, 0.205, 0.35, np.nan], [0.45, 0.55, 0.018, np.nan], [0.74, 1.1, -0.001, 1e-06]])
    df_cols = ['Sg', 'Krg', 'Kro', 'Pc']
    df = pd.DataFrame(np_df, columns = df_cols)
    phase_combo = 'gas-oil'

    with pytest.raises(ValueError) as excval:
        RelPerm(model = model, df = df, phase_combo = phase_combo)
    assert "Krg is not within the range 0-1" in str(excval.value)


def test_relperm(tmp_path):
    epc = os.path.join(tmp_path, 'model.epc')
    model = rq.new_model(epc)
    np_df = np.array([[0.0, 0.0, 1.0, 0], [0.04, 0.015, 0.87, np.nan], [0.12, 0.065, 0.689, np.nan],
                      [0.25, 0.205, 0.35, np.nan], [0.45, 0.55, 0.019, np.nan], [0.74, 1.0, 0.0, 0.000001]])
    df_cols1 = ['Sw', 'Krw', 'Kro', 'Pc']
    df1 = pd.DataFrame(np_df, columns = df_cols1)
    phase_combo1 = 'oil-water'
    uom_list = ['Euc', 'Euc', 'Euc', 'psi']
    df_cols2 = ['Sg', 'Krg', 'Kro', 'Pc']
    df2 = pd.DataFrame(np_df, columns = df_cols2)
    phase_combo2 = 'gas-oil'
    dataframe1 = RelPerm(model = model,
                         df = df1,
                         uom_list = uom_list,
                         phase_combo = phase_combo1,
                         low_sal = True,
                         table_index = 1,
                         title = 'table1')
    dataframe2 = RelPerm(model = model,
                         df = df2,
                         uom_list = uom_list,
                         phase_combo = phase_combo2,
                         low_sal = False,
                         title = 'table2')
    assert dataframe1.n_cols == 4
    assert dataframe1.n_rows == 6
    assert_frame_equal(dataframe1.dataframe(), df1)
    assert all(dataframe1.dataframe().columns == df_cols1)
    assert round(dataframe1.interpolate_point(saturation = 0.55, kr_or_pc_col = 'Kro')[1], 3) == 0.012
    dataframe1.write_hdf5_and_create_xml()
    dataframe2.write_hdf5_and_create_xml()
    assert model.part(extra = {'relperm_table': 'true', 'low_sal': 'true'}) == model.part(title = 'table1')
    assert len(relperm_parts_in_model(model, low_sal = False)) == 1
    dataframe1.df_to_text(filepath = tmp_path, filename = 'oil_water_test_table')
    # reconstruct a dataframe of rel. perm. data from a text file
    df1_reconstructed_from_file = text_to_relperm_dict(os.path.join(tmp_path,
                                                                    'oil_water_test_table.dat'))['relperm_table1']['df']
    assert_frame_equal(df1, df1_reconstructed_from_file)
    assert df1_reconstructed_from_file.iloc[3]['Kro'] == 0.350
    # reconstruct a dataframe of rel. perm. data from a string
    with open(os.path.join(tmp_path, 'oil_water_test_table.dat')) as f:
        relperm_string = f.read()
    df1_reconstructed_from_string = text_to_relperm_dict(relperm_string, is_file = False)['relperm_table1']['df']
    assert_frame_equal(df1, df1_reconstructed_from_string)
    # initialize a RelPerm object from an existing uuid
    new_wo_obj = RelPerm(model, uuid = model.uuid(title = 'table1'))
    assert new_wo_obj.phase_combo == phase_combo1
    assert new_wo_obj.low_sal == 'true'
    assert new_wo_obj.table_index == '1'
    assert new_wo_obj.extra_metadata == {
        'dataframe': 'true',
        'low_sal': 'true',
        'phase_combo': 'oil-water',
        'relperm_table': 'true',
        'table_index': '1'
    }


def test_relperm_df_none_uuid_none(tmp_path):
    # Arrange
    test_model = rq.new_model(os.path.join(tmp_path, 'test'))

    # Act & Assert
    with pytest.raises(ValueError) as e:
        RelPerm(test_model)
    assert "either a uuid or a dataframe must be provided" in str(e.value)


def test_relperm_invalid_phase_combo(tmp_path):
    # Arrange
    test_model = rq.new_model(os.path.join(tmp_path, 'test'))
    np_df = np.array([[0.0, 0.0, 1.0, 0], [0.04, 0.015, 0.87, np.nan], [0.12, 0.065, 0.689, np.nan],
                      [0.25, 0.205, 0.35, np.nan], [0.45, 0.55, 0.019, np.nan], [0.74, 1.0, 0.0, 0.000001]])
    df_cols = ['Sw', 'Krw', 'Kro', 'Pc']
    test_df = pd.DataFrame(np_df, columns = df_cols)
    test_phase_combo = 'wateroil'

    # Act & Assert
    with pytest.raises(ValueError) as e:
        RelPerm(test_model, df = test_df, phase_combo = test_phase_combo)
    assert "invalid phase_combo provided" in str(e.value)


def test_relperm_table_index_equal_0(tmp_path):
    # Arrange
    test_model = rq.new_model(os.path.join(tmp_path, 'test'))
    np_df = np.array([[0.0, 0.0, 1.0, 0], [0.04, 0.015, 0.87, np.nan], [0.12, 0.065, 0.689, np.nan],
                      [0.25, 0.205, 0.35, np.nan], [0.45, 0.55, 0.019, np.nan], [0.74, 1.0, 0.0, 0.000001]])
    df_cols = ['Sw', 'Krw', 'Kro', 'Pc']
    test_df = pd.DataFrame(np_df, columns = df_cols)
    test_table_index = 0

    # Act & Assert
    with pytest.raises(ValueError) as e:
        RelPerm(test_model, df = test_df, table_index = test_table_index)
    assert "table_index cannot be less than 1" in str(e.value)


@pytest.mark.parametrize("test_phase_combo", ['gas-water', 'water-gas'])
def test_relperm_gas_water_phase_combo_invalid_first_column(tmp_path, test_phase_combo):
    # Arrange
    test_model = rq.new_model(os.path.join(tmp_path, 'test'))
    np_df = np.array([[0.0, 0.0, 1.0, 0], [0.04, 0.015, 0.87, np.nan], [0.12, 0.065, 0.689, np.nan],
                      [0.25, 0.205, 0.35, np.nan], [0.45, 0.55, 0.019, np.nan], [0.74, 1.0, 0.0, 0.000001]])
    df_cols = ['So', 'Krw', 'Kro', 'Pc']
    test_df = pd.DataFrame(np_df, columns = df_cols)

    # Act & Assert
    with pytest.raises(ValueError) as e:
        RelPerm(test_model, df = test_df, phase_combo = test_phase_combo)
    assert "incorrect saturation column name and/or multiple saturation columns exist" in str(e.value)


@pytest.mark.parametrize("test_phase_combo", ['gas-water', 'water-gas'])
def test_relperm_gas_water_phase_combo_invalid_columns(tmp_path, test_phase_combo):
    # Arrange
    test_model = rq.new_model(os.path.join(tmp_path, 'test'))
    np_df = np.array([[0.0, 0.0, 1.0, 0], [0.04, 0.015, 0.87, np.nan], [0.12, 0.065, 0.689, np.nan],
                      [0.25, 0.205, 0.35, np.nan], [0.45, 0.55, 0.019, np.nan], [0.74, 1.0, 0.0, 0.000001]])
    df_cols = ['Sg', 'Krw', 'Kro', 'Pc']
    test_df = pd.DataFrame(np_df, columns = df_cols)

    # Act & Assert
    with pytest.raises(ValueError) as e:
        RelPerm(test_model, df = test_df, phase_combo = test_phase_combo)
    assert "incorrect column name(s) {'Kro'}" in str(e.value)


def test_relperm_no_phase_combo_invalid_first_column(tmp_path):
    # Arrange
    test_model = rq.new_model(os.path.join(tmp_path, 'test'))
    np_df = np.array([[0.0, 0.0, 1.0, 0], [0.04, 0.015, 0.87, np.nan], [0.12, 0.065, 0.689, np.nan],
                      [0.25, 0.205, 0.35, np.nan], [0.45, 0.55, 0.019, np.nan], [0.74, 1.0, 0.0, 0.000001]])
    df_cols = ['Sn', 'Krw', 'Kro', 'Pc']
    test_df = pd.DataFrame(np_df, columns = df_cols)

    # Act & Assert
    with pytest.raises(ValueError) as e:
        RelPerm(test_model, df = test_df)
    assert "incorrect saturation column name and/or multiple saturation columns exist" in str(e.value)


@pytest.mark.parametrize("test_cols,test_phase_combo", [(['Sw', 'Krw', 'Kro', 'Pc'], 'water-oil'),
                                                        (['Sg', 'Krg', 'Kro', 'Pc'], 'gas-oil'),
                                                        (['Sw', 'Krw', 'Krg', 'Pc'], 'gas-water')])
def test_relperm_no_phase_combo(tmp_path, test_cols, test_phase_combo):
    # Arrange
    test_model = rq.new_model(os.path.join(tmp_path, 'test'))
    np_df = np.array([[0.0, 0.0, 1.0, 0], [0.04, 0.015, 0.87, np.nan], [0.12, 0.065, 0.689, np.nan],
                      [0.25, 0.205, 0.35, np.nan], [0.45, 0.55, 0.019, np.nan], [0.74, 1.0, 0.0, 0.000001]])
    test_df = pd.DataFrame(np_df, columns = test_cols)

    # Act
    relperm = RelPerm(test_model, df = test_df)

    # Assert
    assert relperm.phase_combo == test_phase_combo



"""The Grid Module"""

__all__ = [
    'Grid', 'transmissibility', 'RegularGrid', 'extract_grid_parent', 'establish_zone_property_kind',
    'find_cell_for_x_sect_xz', 'grid_flavour', 'is_regular_grid', 'any_grid'
]

from .transmissibility import transmissibility
from .grid import Grid
from .regular_grid import RegularGrid
from .grid_types import grid_flavour, is_regular_grid, any_grid
from .moved_functions import establish_zone_property_kind
from .extract_functions import extract_grid_parent, extent_kji_from_root
from .points_functions import find_cell_for_x_sect_xz



"""Submodule containing the RegularGrid class."""

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.crs as rqc
import resqpy.olio.transmission as rqtr
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vec
import resqpy.property as rprop
from .grid import Grid
from .grid_types import is_regular_grid

always_write_pillar_geometry_is_defined_array = False
always_write_cell_geometry_is_defined_array = False


class RegularGrid(Grid):
    """Class for completely regular block grids aligned with xyz axes."""

    # For now generate a standard unsplit pillar grid
    # todo: use RESQML lattice like geometry specification

    def __init__(self,
                 parent_model,
                 extent_kji = None,
                 dxyz = None,
                 dxyz_dkji = None,
                 origin = (0.0, 0.0, 0.0),
                 crs_uuid = None,
                 use_vertical = False,
                 mesh = None,
                 mesh_dz_dk = 1.0,
                 uuid = None,
                 set_points_cached = False,
                 as_irregular_grid = False,
                 find_properties = True,
                 title = None,
                 originator = None,
                 extra_metadata = {}):
        """Creates a regular grid object based on dxyz, or derived from a Mesh object.

        arguments:
           parent_model (model.Model object): the model to which the new grid will be assigned
           extent_kji (triple positive integers, optional): the number of cells in the grid (nk, nj, ni);
              required unless grid_root is present
           dxyz (triple float, optional): use when the I,J,K axes align with the x,y,z axes (possible with inverted
              directions); the size of each cell (dx, dy, dz); values may be negative
           dxyz_dkji (numpy float array of shape (3, 3), optional): how x,y,z values increase with each step in each
              direction K,J,I; first index is KJI, second index is xyz; only one of dxyz, dxyz_dkji and mesh should be
              present; NB axis ordering is different to that used in Mesh class for dxyz_dij
           origin (triple float, default (0.0, 0.0, 0.0)): the location in the local coordinate space of the crs of
              the 'first' corner point of the grid
           crs_uuid (uuid.UUID, optional): the uuid of the coordinate reference system for the grid
           use_vertical (boolean, default False): if True and the pillars of the regular grid are vertical then a
              pillar shape of 'vertical' is used; if False (or the pillars are not vertical), then a pillar shape of
              'straight' is used
           mesh (surface.Mesh object, optional): if present, the I,J layout of the grid is based on the mesh, which
              must be regular, and the K cell size is given by the mesh_dz_dk argument; if present, then dxyz and
              dxyz_dkji must be None
           mesh_dz_dk (float, default 1.0): the size of cells in the K axis, which is aligned with the z axis, when
              starting from a mesh; ignored if mesh is None
           uuid (optional): the root of the xml tree for the grid part; if present, the RegularGrid object is
              based on existing data or a mix of that data and other arguments where present
           set_points_cached (boolean, default False): if True, an explicit geometry is created for the regular grid
              in the form of the cached points array; will be treated as True if as_irregular_grid is True
           as_irregular_grid (boolean, default False): if True, the grid is setup such that it will appear as a Grid
              object when next loaded from disc
           find_properties (boolean, default True): if True and grid_root is not None, a grid property collection is
              instantiated as an attribute, holding properties for which this grid is the supporting representation
           title (str, optional): citation title for new grid; ignored if loading from xml
           originator (str, optional): name of person creating the grid; defaults to login id;
              ignored if loading from xml
           extra_metadata (dict, optional): dictionary of extra metadata items to add to the grid;
              ignored if loading from xml

        returns:
           a newly created RegularGrid object with inheritance from the Grid class

        notes:
           the RESQML standard allows for regular grid geometry pillars to be stored as parametric lines
           but that is not yet supported by this code base; however, constant dx, dy, dz arrays are supported;
           alternatively, regular meshes (Grid2d) may be stored in parameterized form and used to generate a
           regular grid here;
           if root_grid, dxyz, dxyz_dkji and mesh arguments are all None then unit cube cells aligned with
           the x,y,z axes will be generated;
           to store the geometry explicitly use the following methods: make_regular_points_cached(), write_hdf5(),
           create_xml(..., write_geometry = True);
           otherwise, avoid write_hdf5() and call create_xml(..., write_geometry = False)

        :meta common:
        """

        if as_irregular_grid:
            set_points_cached = True

        self.is_aligned = None  #: boolean indicating alignment of IJK axes with +/- xyz respectively

        if uuid is None:
            super().__init__(parent_model, title = title, originator = originator, extra_metadata = extra_metadata)
            self.grid_representation = 'IjkGrid' if as_irregular_grid else 'IjkBlockGrid'
            self.extent_kji = np.array(extent_kji).copy()
            self.nk, self.nj, self.ni = self.extent_kji
            self.k_direction_is_down = True  # assumed direction
            self.grid_is_right_handed = False  # todo: work it out from dxyz_dkji and crs xyz handedness
            self.has_split_coordinate_lines = False
            self.k_gaps = None
            self.k_gap_after_array = None
            self.k_raw_index_array = None
            self.inactive = None
            self.all_inactive = None
            self.geometry_defined_for_all_cells_cached = True
            self.geometry_defined_for_all_pillars_cached = True
            self.array_cell_geometry_is_defined = np.full(tuple(self.extent_kji), True, dtype = bool)
        else:
            assert is_regular_grid(parent_model.root_for_uuid(uuid))
            super().__init__(parent_model,
                             uuid = uuid,
                             find_properties = find_properties,
                             geometry_required = False,
                             title = title,
                             originator = originator,
                             extra_metadata = extra_metadata)
            self.grid_representation = 'IjkBlockGrid'
            if dxyz is None and dxyz_dkji is None:
                # find cell length properties and populate dxyz from those values
                assert self.property_collection is not None
                dxi_part = self.property_collection.singleton(property_kind = 'cell length',
                                                              facet_type = 'direction',
                                                              facet = 'I')
                dyj_part = self.property_collection.singleton(property_kind = 'cell length',
                                                              facet_type = 'direction',
                                                              facet = 'J')
                dzk_part = self.property_collection.singleton(property_kind = 'cell length',
                                                              facet_type = 'direction',
                                                              facet = 'K')
                if dxi_part is not None and dyj_part is not None and dzk_part is not None:
                    dxi = self.property_collection.constant_value_for_part(dxi_part)
                    dyj = self.property_collection.constant_value_for_part(dyj_part)
                    dzk = self.property_collection.constant_value_for_part(dzk_part)
                    if dxi is not None and dyj is not None and dzk is not None:
                        dxyz = (float(dxi), float(dyj), float(dzk))

        if mesh is not None:
            assert mesh.flavour == 'regular'
            assert dxyz is None and dxyz_dkji is None
            origin = mesh.regular_origin
            dxyz_dkji = np.empty((3, 3))
            dxyz_dkji[0, :] = mesh_dz_dk
            dxyz_dkji[1, :] = mesh.regular_dxyz_dij[1]  # J axis
            dxyz_dkji[2, :] = mesh.regular_dxyz_dij[0]  # I axis
            if crs_uuid is None:
                crs_uuid = mesh.crs_uuid
            else:
                assert bu.matching_uuids(crs_uuid, mesh.crs_uuid)

        assert dxyz is None or dxyz_dkji is None
        if dxyz is None and dxyz_dkji is None:
            dxyz = (1.0, 1.0, 1.0)
        if dxyz_dkji is None:
            dxyz_dkji = np.array([[0.0, 0.0, dxyz[2]], [0.0, dxyz[1], 0.0], [dxyz[0], 0.0, 0.0]])
        self.block_origin = np.array(origin).copy()
        self.block_dxyz_dkji = np.array(dxyz_dkji).copy()
        self._set_is_aligned()
        if use_vertical and dxyz_dkji[0][0] == 0.0 and dxyz_dkji[0][1] == 0.0:  # ie. no x,y change with k
            self.pillar_shape = 'vertical'
        else:
            self.pillar_shape = 'straight'

        if set_points_cached:
            self.make_regular_points_cached()

        if crs_uuid is None:
            crs_uuid = parent_model.crs_uuid
        if crs_uuid is None:
            new_crs = rqc.Crs(parent_model)
            self.crs_root = new_crs.create_xml(reuse = True)
            self.crs_uuid = new_crs.uuid
        else:
            self.crs_uuid = crs_uuid
            self.crs_root = parent_model.root_for_uuid(crs_uuid)

        if self.uuid is None:
            self.uuid = bu.new_uuid()

    def make_regular_points_cached(self):
        """Set up the cached points array as an explicit representation of the regular grid geometry."""

        if hasattr(self, 'points_cached') and self.points_cached is not None:
            return
        self.points_cached = np.zeros((self.nk + 1, self.nj + 1, self.ni + 1, 3))
        # todo: replace for loops with linspace
        for k in range(self.nk):
            self.points_cached[k + 1, 0, 0] = self.points_cached[k, 0, 0] + self.block_dxyz_dkji[0]
        for j in range(self.nj):
            self.points_cached[:, j + 1, 0] = self.points_cached[:, j, 0] + self.block_dxyz_dkji[1]
        for i in range(self.ni):
            self.points_cached[:, :, i + 1] = self.points_cached[:, :, i] + self.block_dxyz_dkji[2]
        self.points_cached[:, :, :] += self.block_origin

    def axial_lengths_kji(self):
        """Returns a triple float being lengths of primary axes (K, J, I) for each cell."""

        return vec.naive_lengths(self.block_dxyz_dkji)

    # override of Grid methods

    def point_raw(self, index = None, points_root = None, cache_array = True):
        """Returns element from points data, indexed as corner point (k0, j0, i0).

        Can optionally be used to cache points data.

        arguments:
           index (3 integers, optional): if not None, the index into the raw points data for the point of interest
           points_root (ignored)
           cache_array (boolean, default True): if True, the raw points data is cached in memory as a side effect

        returns:
           (x, y, z) of selected point as a 3 element numpy vector, or None if index is None

        notes:
           this function is typically called either to cache the points data in memory, or to fetch the coordinates of
           a single corner point;
           the index should be a triple kji0 with axes ranging over the shared corners nk+1, nj+1, ni+1
        """

        assert cache_array or index is not None

        if cache_array:
            self.make_regular_points_cached()
            if index is None:
                return None
            return self.points_cached[tuple(index)]

        return self.block_origin + np.sum(np.repeat(np.array(index).reshape(
            (3, 1)), 3, axis = -1) * self.block_dxyz_dkji,
                                          axis = 0)

    def half_cell_transmissibility(self, use_property = None, realization = None, tolerance = None):
        """Returns (and caches if realization is None) half cell transmissibilities for this regular grid.

        arguments:
           use_property (ignored)
           realization (int, optional) if present, only a property with this realization number will be used
           tolerance (ignored)

        returns:
           numpy float array of shape (nk, nj, ni, 3, 2) where the 3 covers K,J,I and the 2 covers the
              face polarity: - (0) and + (1); units will depend on the length units of the coordinate reference
              system for the grid; the units will be m3.cP/(kPa.d) or bbl.cP/(psi.d) for grid length units of m
              and ft respectively

        notes:
           the values for - and + polarity will always be equal, the data is duplicated for compatibility with
           parent Grid class;
           the returned array is in the logical resqpy arrangement; it must be discombobulated before being
           added as a property; this method does not write to hdf5, nor create a new property or xml;
           if realization is None, a grid attribute cached array will be used
        """

        # todo: allow passing of property uuids for ntg, k_k, j, i

        if realization is None and hasattr(self, 'array_half_cell_t'):
            return self.array_half_cell_t

        half_t = rqtr.half_cell_t(
            self, realization = realization)  # note: properties must be identifiable in property_collection

        # introduce facial polarity axis for compatibility with parent Grid class
        assert half_t.ndim == 4
        half_t = np.expand_dims(half_t, -1)
        half_t = np.repeat(half_t, 2, axis = -1)

        if realization is None:
            self.array_half_cell_t = half_t

        return half_t

    def centre_point(self, cell_kji0 = None, cache_centre_array = False):
        """Returns centre point of a cell or array of centre points of all cells.

        arguments:
           cell_kji0 (optional): if present, the (k, j, i) indices of the individual cell for which the
              centre point is required; zero based indexing
           cache_centre_array (boolean, default False): If True, or cell_kji0 is None, an array of centre points
              is generated and added as an attribute of the grid, with attribute name array_centre_point

        returns:
           (x, y, z) 3 element numpy array of floats holding centre point of cell;
           or numpy 3+1D array if cell_kji0 is None

        note:
           resulting coordinates are in the same (local) crs as the grid points
        """

        if cell_kji0 is None:
            cache_centre_array = True

        if cache_centre_array and (not hasattr(self, 'array_centre_point') or self.array_centre_point is None):
            centres = np.zeros((self.nk, self.nj, self.ni, 3))
            # todo: replace for loops with linspace
            for k in range(self.nk - 1):
                centres[k + 1, 0, 0] = centres[k, 0, 0] + self.block_dxyz_dkji[0]
            for j in range(self.nj - 1):
                centres[:, j + 1, 0] = centres[:, j, 0] + self.block_dxyz_dkji[1]
            for i in range(self.ni - 1):
                centres[:, :, i + 1] = centres[:, :, i] + self.block_dxyz_dkji[2]
            centres += self.block_origin + 0.5 * np.sum(self.block_dxyz_dkji, axis = 0)
            self.array_centre_point = centres

        if cell_kji0 is not None:
            if hasattr(self, 'array_centre_point') and self.array_centre_point is not None:
                return self.array_centre_point[tuple(cell_kji0)]
            float_kji0 = np.array(cell_kji0, dtype = float) + 0.5
            centre = self.block_origin + np.sum(
                self.block_dxyz_dkji * np.expand_dims(float_kji0, axis = -1).repeat(3, axis = -1), axis = 0)
            return centre

        return self.array_centre_point

    def volume(self, cell_kji0 = None):
        """Returns bulk rock volume of cell or numpy array of bulk rock volumes for all cells.

        arguments:
           cell_kji0 (optional): if present, the (k, j, i) indices of the individual cell for which the
                                 volume is required; zero based indexing

        returns:
           float, being the volume of cell identified by cell_kji0;
           or numpy float array of shape (nk, nj, ni) if cell_kji0 is None

        notes:
           the function can be used to find the volume of a single cell, or all cells;
           grid's coordinate reference system must use same units in z as xy (projected);
           units of result are implicitly determined by coordinates in grid's coordinate reference system;
           the method currently assumes that the primary i, j, k axes are mutually orthogonal
        """

        vol = np.product(vec.naive_lengths(self.block_dxyz_dkji))
        if cell_kji0 is not None:
            return vol
        return np.full((self.nk, self.nj, self.ni), vol)

    def thickness(self, cell_kji0 = None, **kwargs):
        """Returns cell thickness (K axial length) for a single cell or full array.

        arguments:
           cell_kji0 (triple int, optional): if present, the thickness for a single cell is returned;
              if None, an array is returned
           all other arguments ignored; present for compatibility with same method in Grid()

        returns:
           float, or numpy float array filled with a constant
        """

        thick = self.axial_lengths_kji()[0]
        if cell_kji0 is not None:
            return thick
        return np.full((self.nk, self.nj, self.ni), thick, dtype = float)

    def pinched_out(self, cell_kji0 = None, **kwargs):
        """Returns pinched out boolean (always False) for a single cell or full array.

        arguments:
           cell_kji0 (triple int, optional): if present, the pinched out flag for a single cell is returned;
              if None, an array is returned
           all other arguments ignored; present for compatibility with same method in Grid()

        returns:
           False, or numpy array filled with False
        """

        if cell_kji0 is not None:
            return False
        return np.full((self.nk, self.nj, self.ni), False, dtype = bool)

    def actual_pillar_shape(self, patch_metadata = False, tolerance = 0.001):
        """Returns actual shape of pillars.

        arguments:
           patch_metadata (boolean, default False): if True, the actual shape replaces whatever was in the metadata
           tolerance (float, ignored)

        returns:
           string: 'vertical', 'straight' or 'curved'

        note:
           setting patch_metadata True will affect the attribute in this Grid object; however, it will not be
           preserved unless the create_xml() method is called, followed at some point with model.store_epc()
        """

        if np.all(self.block_dxyz_dkji[0, :2] == 0.0):
            return 'vertical'
        return 'straight'

    def create_xml(self,
                   ext_uuid = None,
                   add_as_part = True,
                   add_relationships = True,
                   set_as_grid_root = True,
                   root = None,
                   title = None,
                   originator = None,
                   write_active = True,
                   write_geometry = None,
                   extra_metadata = {},
                   expand_const_arrays = False,
                   add_cell_length_properties = True):
        """Creates xml for this RegularGrid object; by default the explicit geometry is not included.

        see docstring for Grid.create_xml()

        additional argument:
           add_cell_length_properties (boolean, default True): if True, 3 constant property arrays with cells as
              indexable element are created to hold the lengths of the primary axes of the cells; the xml is
              created for the properties and they are added to the model (no hdf5 write needed)

        :meta common:
        """

        if write_geometry is None:
            write_geometry = (self.grid_representation == 'IjkGrid')

        node = super().create_xml(ext_uuid = ext_uuid,
                                  add_as_part = add_as_part,
                                  add_relationships = add_relationships,
                                  set_as_grid_root = set_as_grid_root,
                                  title = title,
                                  originator = originator,
                                  write_active = write_active,
                                  write_geometry = write_geometry,
                                  extra_metadata = extra_metadata)

        if add_cell_length_properties:
            axes_lengths_kji = self.axial_lengths_kji()
            dpc = rprop.GridPropertyCollection()
            dpc.set_grid(self)
            for axis in range(3):
                dpc.add_cached_array_to_imported_list(None,
                                                      'regular grid',
                                                      'D' + 'ZYX'[axis],
                                                      discrete = False,
                                                      uom = self.xy_units(),
                                                      property_kind = 'cell length',
                                                      facet_type = 'direction',
                                                      facet = 'KJI'[axis],
                                                      indexable_element = 'cells',
                                                      count = 1,
                                                      const_value = axes_lengths_kji[axis])
            if expand_const_arrays:
                dpc.write_hdf5_for_imported_list(expand_const_arrays = True)
            dpc.create_xml_for_imported_list_and_add_parts_to_model(expand_const_arrays = expand_const_arrays)
            if self.property_collection is None:
                self.property_collection = dpc
            else:
                if self.property_collection.support is None:
                    self.property_collection.set_support(support = self)
                self.property_collection.inherit_parts_from_other_collection(dpc)

        return node

    def _set_is_aligned(self):
        """Sets is_aligned attribute True if IJK axes align with +/- xyz respectively."""
        if self.block_dxyz_dkji is None:
            self.is_aligned = None
        self.is_aligned = (np.count_nonzero(self.block_dxyz_dkji * np.array([[1.0, 1, 0], [1, 0, 1], [0, 1, 1]])) == 0)



import numpy as np
import pytest
import resqpy.olio.utility as u


@pytest.mark.parametrize("test_input,expected_output",
                         [(np.array([2, 5, 3]), np.array([2, 4, 1])), (np.array([-3, 0, -6]), np.array([-7, -1, -4])),
                          (np.array([-193725, 481873, -330798]), np.array([-330799, 481872, -193726]))])
def test_kji0_from_ijk1(test_input, expected_output):
    # Act
    output = u.kji0_from_ijk1(test_input)

    # Assert
    np.testing.assert_array_equal(output, expected_output)


@pytest.mark.parametrize("test_input,expected_output",
                         [(np.array([2, 4, 1]), np.array([2, 5, 3])), (np.array([-7, -1, -4]), np.array([-3, 0, -6])),
                          (np.array([-330799, 481872, -193726]), np.array([-193725, 481873, -330798]))])
def test_ijk1_from_kji0(test_input, expected_output):
    # Act
    output = u.ijk1_from_kji0(test_input)

    # Assert
    np.testing.assert_array_equal(output, expected_output)


@pytest.mark.parametrize("test_input,expected_output",
                         [(np.array([2, 5, 3]), np.array([3, 5, 2])), (np.array([-3, 0, -6]), np.array([-6, 0, -3])),
                          (np.array([-193725, 481873, -330798]), np.array([-330798, 481873, -193725]))])
def test_extent_switch_ijk_kji(test_input, expected_output):
    # Act
    output = u.extent_switch_ijk_kji(test_input)

    # Assert
    np.testing.assert_array_equal(output, expected_output)


@pytest.mark.parametrize("test_input,expected_output", [(np.array([2, 5, 3]), 30), ([3, 0, 6], 0),
                                                        ((193, 481, 330), 30634890)])
def test_cell_count_from_extent(test_input, expected_output):
    # Act
    output = u.cell_count_from_extent(test_input)

    # Assert
    assert output == expected_output


@pytest.mark.parametrize("test_input,expected_output",
                         [(np.array([2, 5, 3]), '[4, 6, 3]'), (np.array([-3, 0, -6]), '[-5, 1, -2]'),
                          (np.array([-193725, 481873, -330798]), '[-330797, 481874, -193724]')])
def test_string_ijk1_for_cell_kji0(test_input, expected_output):
    # Act
    output = u.string_ijk1_for_cell_kji0(test_input)

    # Assert
    assert output == expected_output


@pytest.mark.parametrize("test_input,expected_output",
                         [(np.array([2, 5, 3]), '[2, 5, 3]'), (np.array([-3, 0, -6]), '[-3, 0, -6]'),
                          (np.array([-193725, 481873, -330798]), '[-193725, 481873, -330798]')])
def test_string_ijk1_for_cell_ijk1(test_input, expected_output):
    # Act
    output = u.string_ijk1_for_cell_ijk1(test_input)

    # Assert
    assert output == expected_output


@pytest.mark.parametrize("test_input,expected_output",
                         [(np.array([2, 5, 3]), '[3, 5, 2]'), (np.array([-3, 0, -6]), '[-6, 0, -3]'),
                          (np.array([-193725, 481873, -330798]), '[-330798, 481873, -193725]')])
def test_string_ijk_for_extent_kji(test_input, expected_output):
    # Act
    output = u.string_ijk_for_extent_kji(test_input)

    # Assert
    assert output == expected_output


@pytest.mark.parametrize("test_input,expected_output", [(np.array([26.1, 5.0, 32.8]), '(26.10, 5.00, 32.800)'),
                                                        ([327.281, 82.19, 62172.329], '(327.28, 82.19, 62172.329)'),
                                                        ((19323.1, 481.33, 330.201), '(19323.10, 481.33, 330.201)')])
def test_string_xyz(test_input, expected_output):
    # Act
    output = u.string_xyz(test_input)

    # Assert
    assert output == expected_output


@pytest.mark.parametrize("test_input1,test_input2,expected_output", [(3, '+', 3.5), (-1, '-', -1.5), (7, '-', 6.5)])
def test_horizon_float(test_input1, test_input2, expected_output):
    # Act
    output = u.horizon_float(test_input1, test_input2)

    # Assert
    assert output == expected_output



"""RESQML manipulation library.

.. autosummary::
    :toctree: _autosummary
    :caption: API Reference
    :template: custom-module-template.rst
    :recursive:

    model
    crs
    derived_model
    fault
    grid
    grid_surface
    lines
    organize
    property
    rq_import
    strata
    surface
    time_series
    unstructured
    weights_and_measures
    well
    olio
"""

import logging
import warnings


def _determine_version():

    # Prod setup: Look for hard-coded file resqpy/version.py
    try:
        from .version import version  # type: ignore
        return version
    except Exception:
        pass

    # Dev setup: Use local git history
    try:
        from setuptools_scm import get_version
        return get_version(root = '..', relative_to = __file__)
    except Exception as e:
        warnings.warn("Unable to determine resqpy version: " + str(e))
        return "0.0.0-version-not-available"


__version__ = _determine_version()
log = logging.getLogger(__name__)
log.info(f"Imported resqpy version {__version__}")



import os
import numpy as np
import pytest

from resqpy.olio.grid_functions import left_right_foursome, infill_block_geometry, resequence_nexus_corp, random_cell, determine_corp_ijk_handedness, determine_corp_extent, translate_corp, triangles_for_cell_faces, actual_pillar_shape, columns_to_nearest_split_face
import resqpy.model as rq
import resqpy.grid as grr
from resqpy.lines import Polyline
from resqpy.derived_model import add_faults
from resqpy.olio.random_seed import seed


def test_infill_block_geometry():

    # --------- Arrange ----------
    x_ = np.array([2, 4, 6, 8])
    y_ = np.array([1, 3, 5, 7])
    depth_ = np.array([10.0, 30.0, 50.0, 70.0])
    x, y, depth = np.meshgrid(x_, y_, depth_)
    extent = (4, 4, 4)
    thickness = np.full(extent, 15)

    # Set two inactive cells
    depth[1, 2, 1] = 0.001  # below the default depth_zero_tolerance so should be set to 0 by function
    depth[1, 2, 2] = 0.001
    # Assertions to check that depth does have 2 values == 0.001
    assert np.isclose(depth[1, 2, 1], 0.001)
    assert np.isclose(depth[1, 2, 2], 0.001)
    np.testing.assert_almost_equal(np.unique(thickness), np.array([15.]))
    # --------- Act ----------
    infill_block_geometry(extent = extent, depth = depth.transpose(), thickness = thickness, x = x, y = y)
    # --------- Assert ----------
    assert len(depth[np.isclose(depth, 0.)]) == 0
    assert len(thickness[np.isclose(thickness, 22.)]) == 2


def test_infill_block_geometry_log_warnings(caplog):

    # --------- Arrange ----------
    x_ = np.array([2, 4, 6, 8])
    y_ = np.array([1, 3, 5, 7])
    depth_ = np.array([10.0, 30.0, 50.0, 70.0])
    x, y, depth = np.meshgrid(x_, y_, depth_)
    extent = (4, 4, 4)
    thickness = np.ones(extent) * 15

    # Set two inactive cells
    depth[1, 2, 1] = 0.001  # below the default depth_zero_tolerance so should be set to 0 by function
    depth[1, 2, 2] = 0.001
    # Assertions to check that depth does have 2 values == 0.001
    assert np.isclose(depth[1, 2, 1], 0.001)
    assert np.isclose(depth[1, 2, 2], 0.001)
    np.testing.assert_almost_equal(np.unique(thickness), np.array([15.]))
    # --------- Act ----------
    infill_block_geometry(extent = extent, depth = depth, thickness = thickness, x = x, y = y)
    # --------- Assert ----------
    assert 'Check k_increase_direction and tolerances' in caplog.text
    assert 'Cells [3, 3, 1] and [3, 3, 3] overlap ...' in caplog.text
    assert 'Skipping rest of i,j column' in caplog.text


@pytest.mark.parametrize('previous, next, expected_result', [
    ((1, 0), (0, 1), np.array([[False, True], [True, True]], dtype = bool)),
    ((1, 0), (1, 2), np.array([[False, False], [True, True]], dtype = bool)),
    ((1, 0), (2, 1), np.array([[False, False], [True, False]], dtype = bool)),
    ((1, 2), (0, 1), np.array([[False, True], [False, False]], dtype = bool)),
    ((1, 2), (1, 0), np.array([[True, True], [False, False]], dtype = bool)),
    ((1, 2), (2, 1), np.array([[True, True], [True, False]], dtype = bool)),
    ((0, 1), (1, 0), np.array([[True, False], [False, False]], dtype = bool)),
    ((0, 1), (2, 1), np.array([[True, False], [True, False]], dtype = bool)),
    ((0, 1), (1, 2), np.array([[True, False], [True, True]], dtype = bool)),
    ((2, 1), (1, 0), np.array([[True, True], [False, True]], dtype = bool)),
    ((2, 1), (0, 1), np.array([[False, True], [False, True]], dtype = bool)),
    ((2, 1), (1, 2), np.array([[False, False], [False, True]], dtype = bool)),
])
def test_left_right_foursome(previous, next, expected_result):

    # --------- Arrange ----------
    full_p_list = [previous, (1, 1), next]
    p_index = 1
    # --------- Act ----------
    result = left_right_foursome(full_pillar_list = full_p_list, p_index = p_index)
    # --------- Assert ----------
    np.testing.assert_equal(result, expected_result)


def test_left_right_foursome_error_handling():

    # --------- Arrange ----------
    p_index = 1
    # arrange for exceptions
    full_p_list_entry_error = [(1, 0), (1, 2), (0, 1)]
    full_p_list_exit_error = [(1, 0), (1, 1), (2, 2)]

    # --------- Act and Assert----------
    with pytest.raises(Exception) as excinfo:
        result_entry_error = left_right_foursome(full_pillar_list = full_p_list_entry_error, p_index = p_index)
        assert "code failure whilst taking entry sides from dubious full pillar list" in str(excinfo.value)
    with pytest.raises(Exception) as excinfo2:
        result_exit_error = left_right_foursome(full_pillar_list = full_p_list_exit_error, p_index = p_index)
        assert "code failure whilst taking exit sides from dubious full pillar list" in str(excinfo.value)


def test_resequence_nexus_corp():

    # --------- Arrange ----------
    corner_points = np.array([[[[[[[0., 0., 0.], [10., 0., 0.]], [[10., -10., 0.], [0., -10., 0.]]],
                                 [[[0., 0., 10.], [10., 0., 10.]], [[10., -10., 10.], [0., -10., 10.]]]]]]])

    expected_result = np.array([[[[[[[0., 0., 0.], [10., 0., 0.]], [[0., -10., 0.], [10., -10., 0.]]],
                                   [[[0., 0., 10.], [10., 0., 10.]], [[0., -10., 10.], [10., -10., 10.]]]]]]])

    # --------- Act ----------
    resequence_nexus_corp(corner_points = corner_points)

    # --------- Assert ----------
    np.testing.assert_almost_equal(corner_points, expected_result)


def test_random_cell(tmp_path):

    # --------- Arrange----------
    seed(1923877)
    epc = os.path.join(tmp_path, 'grid.epc')
    model = rq.new_model(epc)

    # create a basic block grid
    dxyz = (10.0, 10.0, 100.0)
    grid = grr.RegularGrid(model,
                           extent_kji = (4, 2, 2),
                           title = 'grid_1',
                           origin = (0.0, 10.0, 1000.0),
                           dxyz = dxyz,
                           as_irregular_grid = True)

    grid_points = grid.points_ref(masked = False)

    # pinch out cells in the k == 2 layer
    grid_points[3] = grid_points[2]

    # collapse cell kji0 == 0,0,0 in the j direction
    grid_points[0:2, 1, 0:2, 1] = 10.  # same as origin y value

    # store grid
    grid.write_hdf5()
    grid.create_xml(add_cell_length_properties = True)
    grid_uuid = grid.uuid
    model.store_epc()

    # check that the grid can be read
    model = rq.Model(epc)
    grid_reloaded = grr.any_grid(model, uuid = grid_uuid)
    corner_points = grid_reloaded.corner_points()

    # --------- Act----------
    # call random_cell function 50 times
    trial_number = 0
    while trial_number < 50:
        (k, j, i) = random_cell(corner_points = corner_points, border = 0.0)
        # --------- Assert----------
        assert 0 <= k < 4
        assert 0 <= j < 2
        assert 0 <= i < 2
        # check that none of the k,j,i combinations that correspond to pinched cells are chosen by the random_cell function
        assert (k, j, i) not in [(0, 0, 0), (2, 0, 0), (2, 0, 1), (2, 1, 0), (2, 1, 1)]
        trial_number += 1

    # reshape corner get the extent of the new grid
    corner_points_reshaped = corner_points.reshape(1, 1, 16, 2, 2, 2, 3)
    new_extent = determine_corp_extent(corner_points = corner_points_reshaped)
    assert np.all(new_extent == np.array([4, 2, 2], dtype = int))


def test_triangles_for_cell_faces():
    # --------- Arrange----------
    corner_points = np.array([[[[0., 0., 0.], [10., 0., 0.]], [[0., -10., 0.], [10., -10., 0.]]],
                              [[[0., 0., 10.], [10., 0., 10.]], [[0., -10., 10.], [10., -10., 10.]]]])
    # --------- Act----------
    tri = triangles_for_cell_faces(cp = corner_points)
    # face centre points
    k_face_centre_points = [np.array([5., -5., 0.]), np.array([5., -5., 10])]
    j_face_centre_points = [np.array([5., 0., 5.]), np.array([5., -10., 5.])]
    i_face_centre_points = [np.array([0., -5., 5.]), np.array([10., -5., 5.])]

    # --------- Assert----------
    np.testing.assert_almost_equal(tri[0, :, :, 0][0][0], k_face_centre_points[0])
    np.testing.assert_almost_equal(tri[0, :, :, 0][1][0], k_face_centre_points[1])
    np.testing.assert_almost_equal(tri[1, :, :, 0][0][0], j_face_centre_points[0])
    np.testing.assert_almost_equal(tri[1, :, :, 0][1][0], j_face_centre_points[1])
    np.testing.assert_almost_equal(tri[2, :, :, 0][0][0], i_face_centre_points[0])
    np.testing.assert_almost_equal(tri[2, :, :, 0][1][0], i_face_centre_points[1])
    assert tri.shape == (3, 2, 4, 3, 3)


def test_actual_pillar_shape(tmp_path):

    # --------- Arrange----------
    epc = os.path.join(tmp_path, 'grid.epc')
    model = rq.new_model(epc)

    # create a basic block grid
    dxyz = (10.0, 10.0, 10.0)
    vertical_grid = grr.RegularGrid(model,
                                    extent_kji = (2, 2, 2),
                                    title = 'vert_grid',
                                    origin = (0.0, 0.0, 0.0),
                                    dxyz = dxyz)
    straight_grid = grr.RegularGrid(model,
                                    extent_kji = (2, 2, 2),
                                    title = 'straight_grid',
                                    origin = (10.0, 10.0, 0.0),
                                    dxyz = dxyz,
                                    as_irregular_grid = True)
    curved_grid = grr.RegularGrid(model,
                                  extent_kji = (3, 2, 2),
                                  title = 'curved_grid',
                                  origin = (10.0, 10.0, 0.0),
                                  dxyz = dxyz,
                                  as_irregular_grid = True)

    # shift the corner points of cellkji0 == (0, 0, 0) by - 10 x and -10 y units
    straight_grid.corner_points()[0][0][0][0][0][0] = np.array([0., 0., 0])
    straight_grid.corner_points()[0][0][0][0][0][1] = np.array([10., 0., 0])

    # shift 2 corner points of cellkji0 == (1, 0, 0) by -5 x units
    curved_grid.corner_points()[0][0][0][1][0][0] = np.array([5., 10., 10])
    curved_grid.corner_points()[0][0][0][1][0][1] = np.array([15., 10., 10])

    # --------- Act----------
    pillar_shape_vertical = actual_pillar_shape(pillar_points = vertical_grid.corner_points())
    pillar_shape_straight = actual_pillar_shape(pillar_points = straight_grid.corner_points())
    pillar_shape_curved = actual_pillar_shape(pillar_points = curved_grid.corner_points())

    # --------- Assert----------
    assert pillar_shape_vertical == 'vertical'
    assert pillar_shape_straight == 'straight'
    assert pillar_shape_curved == 'curved'


def test_determine_corp_ijk_handedness():

    # --------- Arrange----------
    corner_points_right = np.array([[[[[[[0., 0., 0.], [10., 0., 0.]], [[0., -10., 0.], [10., -10., 0.]]],
                                       [[[0., 0., 10.], [10., 0., 10.]], [[0., -10., 10.], [10., -10., 10.]]]]]]])

    corner_points_left = np.array([[[[[[[0., 0., 0.], [10., 0., 0.]], [[0., 10., 0.], [10., 10., 0.]]],
                                      [[[0., 0., 10.], [10., 0., 10.]], [[0., 10., 10.], [10., 10., 10.]]]]]]])

    # --------- Act----------
    handedness_right = determine_corp_ijk_handedness(corner_points = corner_points_right)
    handedness_left = determine_corp_ijk_handedness(corner_points = corner_points_left)

    # --------- Assert----------
    assert handedness_right == 'right'
    assert handedness_left == 'left'


def test_determine_corp_extent(tmp_path):

    # --------- Arrange----------
    epc = os.path.join(tmp_path, 'grid.epc')
    model = rq.new_model(epc)

    # create a basic block grid
    dxyz = (55.0, 65.0, 27.0)
    grid = grr.RegularGrid(model, extent_kji = (4, 3, 2), title = 'concrete', origin = (0.0, 0.0, 1000.0), dxyz = dxyz)
    grid.create_xml(add_cell_length_properties = True)
    corner_points = grid.corner_points()
    corner_points_reshaped = corner_points.reshape(1, 1, 24, 2, 2, 2, 3)
    # --------- Act----------
    [nk, nj, ni] = determine_corp_extent(corner_points = corner_points_reshaped)

    # --------- Assert----------
    assert [nk, nj, ni] == [4, 3, 2]


@pytest.mark.parametrize(
    # yapf: disable
    'x_shift, y_shift, expected_result',
    [(2, 0,
      np.array([[[[[[[2.001, 0.001, 0.001], [12.001, 0.001, 0.001]], [[2.001, -10.001, 0.001], [12.001, -10.001, 0.001]]
                   ],
                   [[[2.001, 0.001, 10.001], [12.001, 0.001, 10.001]],
                    [[2.001, -10.001, 10.001], [12.001, -10.001, 10.001]]]]]]])),
     (0, 2,
      np.array([[[[[[[0.001, 2.001, 0.001], [10.001, 2.001, 0.001]], [[0.001, -8.001, 0.001], [10.001, -8.001, 0.001]]],
                   [[[0.001, 2.001, 10.001], [10.001, 2.001, 10.001]],
                    [[0.001, -8.001, 10.001], [10.001, -8.001, 10.001]]]]]]]))])
# yapf: enable
def test_translate_corp(x_shift, y_shift, expected_result):

    # --------- Arrange----------
    # yapf: disable
    corner_points = np.array([[[[[[[0.001,   0.001,  0.001], [10.001,   0.001,  0.001]],
                                  [[0.001, -10.001,  0.001], [10.001, -10.001,  0.001]]],
                                 [[[0.001,   0.001, 10.001], [10.001,   0.001, 10.001]],
                                  [[0.001, -10.001, 10.001], [10.001, -10.001, 10.001]]]]]]])
    # yapf: enable

    # --------- Act----------
    translate_corp(corner_points = corner_points, x_shift = x_shift, y_shift = y_shift, preserve_digits = -3)

    # --------- Assert----------
    np.testing.assert_almost_equal(expected_result, corner_points)



import pytest

import resqpy.olio.uuid as uuid
import resqpy.organize as rqo
from resqpy.model import Model


def test_is_equivalent_no_other(tmp_model):
    #Arrange
    fault_interp = rqo.FaultInterpretation(
        tmp_model,
        title = 'test_fault_interpretation',
        domain = "depth",
        is_normal = True,
        maximum_throw = 3,
    )
    #Act
    result = fault_interp.is_equivalent(other = None)

    #Assert
    assert result is False


def test_is_equivalent_other_not_FaultInterpretation(tmp_model):
    # Arrange
    fault_interp = rqo.FaultInterpretation(
        tmp_model,
        title = 'test_fault_interpretation',
        domain = "depth",
        is_normal = True,
        maximum_throw = 3,
    )
    # Act
    result = fault_interp.is_equivalent(other = 55)

    # Assert
    assert result is False


def test_is_equivalent_is_other(tmp_model):
    # Arrange
    fault_interp = rqo.FaultInterpretation(
        tmp_model,
        title = 'test_fault_interpretation',
    )
    # Act
    other = fault_interp
    result = fault_interp.is_equivalent(other = other)

    # Assert
    assert result is True


def test_is_equivalent_is_sameUUIDs(tmp_model):

    # Arrange
    object_uuid = uuid.new_uuid()
    fault_interp = rqo.FaultInterpretation(
        tmp_model,
        title = 'test_fault_interpretation',
    )
    fault_interp.uuid = object_uuid

    # Act
    other = rqo.FaultInterpretation(
        tmp_model,
        title = 'test_fault_interpretation_other',
    )
    other.uuid = object_uuid
    result = fault_interp.is_equivalent(other = other)

    # Assert
    assert result is True


def test_is_equivalent_is_tectonic_boundary_one_none(tmp_model):

    # Arrange
    fault_interp = rqo.FaultInterpretation(
        tmp_model,
        title = 'tectonic_boundary_test',
    )
    fault_interp.tectonic_boundary_feature = 'something'

    # Act
    other = rqo.FaultInterpretation(
        tmp_model,
        title = 'tectonic_boundary_test',
    )
    other.tectonic_boundary_feature = None
    result = fault_interp.is_equivalent(other = other)

    # Assert
    assert result is False


def test_is_equivalent_non_equivalent_tectonic_boundary(tmp_model):

    # Arrange
    tect_boundary = rqo.TectonicBoundaryFeature(tmp_model, kind = 'fault')

    fault_interp = rqo.FaultInterpretation(
        tmp_model,
        title = 'test_fault_interpretation',
    )
    fault_interp.tectonic_boundary_feature = tect_boundary

    # Act
    tect_boundary_other = rqo.TectonicBoundaryFeature(tmp_model, kind = 'fracture')

    other = rqo.FaultInterpretation(
        tmp_model,
        title = 'test_fault_interpretation',
    )
    other.tectonic_boundary_feature = tect_boundary_other
    result = fault_interp.is_equivalent(other = other)

    # Assert
    assert result is False


def test_is_equivalent_throw_math_is_close(tmp_model):

    # Arrange
    fault_interp = rqo.FaultInterpretation(
        tmp_model,
        title = 'test_fault_interpretation',
    )
    fault_interp.maximum_throw = 1

    # Act

    other = rqo.FaultInterpretation(
        tmp_model,
        title = 'test_fault_interpretation',
    )
    other.maximum_throw = 1.002

    result = fault_interp.is_equivalent(other = other)

    # Assert
    assert result is False



"""triangulation.py: functions for finding Delaunay triangulation and Voronoi graph from a set of points."""

version = '5th October 2021'

import logging

log = logging.getLogger(__name__)

import numpy as np
import numpy.typing as npt
from typing import Tuple, List

import resqpy.crs as rqc
import resqpy.lines as rql
import resqpy.model as rq
import resqpy.olio.intersection as meet
import resqpy.olio.vector_utilities as vec

# _ccw_t() no longer needed: triangle vertices maintained in anti-clockwise order throughout
# def _ccw_t(p, t):   # puts triangle vertex indices into anti-clockwise order, in situ
#    if vec.clockwise(p[t[0]], p[t[1]], p[t[2]]) > 0.0:
#       t[1], t[2] = t[2], t[1]


def _dt_simple(po, plot_fn = None, progress_fn = None, container_size_factor = None):
    # returns Delauney triangulation of po and list of hull point indices, using a simple algorithm

    def flip(ei):
        nonlocal fm, e, t, te, p, nt, p_i, ne
        if fm[ei]:
            return  # this edge has already been flipped since last point insertion
        t0, te0 = e[ei, 0]
        t1, te1 = e[ei, 1]
        if t0 < 0 or t1 < 0:
            return  # no triangle on other side
        t0n = te0 - 1
        if t0n < 0:
            t0n = 2
        t1n = te1 - 1
        if t1n < 0:
            t1n = 2
        if (not vec.in_circumcircle(p[t[t0, 0]], p[t[t0, 1]], p[t[t0, 2]], p[t[t1, t1n]]) and
                not vec.in_circumcircle(p[t[t1, 0]], p[t[t1, 1]], p[t[t1, 2]], p[t[t0, t0n]])):
            return  # not sure both are needed
        # flip needed
        ft0 = (t[t0, te0], t[t1, t1n], t[t0, t0n])
        ft1 = (t[t1, te1], t[t0, t0n], t[t1, t1n])
        e[ei, :, 1] = 1
        fte0 = (te[t1, 3 - (te1 + t1n)], ei, te[t0, t0n])
        fte1 = (te[t0, 3 - (te0 + t0n)], ei, te[t1, t1n])
        if e[fte0[0], 0, 0] == t1:
            e[fte0[0], 0] = (t0, 0)
        elif e[fte0[0], 1, 0] == t1:
            e[fte0[0], 1] = (t0, 0)
        else:
            raise Exception('edge breakdown')
        if e[fte1[0], 0, 0] == t0:
            e[fte1[0], 0] = (t1, 0)
        elif e[fte1[0], 1, 0] == t0:
            e[fte1[0], 1] = (t1, 0)
        else:
            raise Exception('edge breakdown')
        if e[fte0[2], 0, 0] == t0:
            e[fte0[2], 0, 1] = 2
        elif e[fte0[2], 1, 0] == t0:
            e[fte0[2], 1, 1] = 2
        else:
            raise Exception('edge breakdown')
        if e[fte1[2], 0, 0] == t1:
            e[fte1[2], 0, 1] = 2
        elif e[fte1[2], 1, 0] == t1:
            e[fte1[2], 1, 1] = 2
        else:
            raise Exception('edge breakdown')
        t[t0] = ft0
        t[t1] = ft1
        te[t0] = fte0
        te[t1] = fte1
        fm[ei] = True
        # debug plot here
        if plot_fn is not None:
            plot_fn(p, t[:nt])
        # recursively flip, not sure all these are needed
        flip(fte0[0])
        flip(fte0[2])
        flip(fte1[0])
        flip(fte1[2])

    n_p = len(po)
    if n_p < 3:
        return None, None  # not enough points
    elif n_p == 3:
        return np.array([0, 1, 2], dtype = int).reshape((1, 3)), np.array([0, 1, 2], dtype = int)

    if progress_fn is not None:
        progress_fn(0.0)

    min_xy = np.min(po[:, :2], axis = 0)
    max_xy = np.max(po[:, :2], axis = 0)
    dxy = max_xy - min_xy
    assert dxy[0] > 0.0 and dxy[1] > 0.0, 'points lie in straight line or are conincident'
    p = np.empty((n_p + 3, 2))
    p[:-3] = po[:, :2]
    # add 3 points sure of containing all po
    csf = 100.0 if container_size_factor is None else max(1.0, container_size_factor)
    p[-3] = (min_xy[0] - 0.8 * csf * dxy[0], min_xy[1] - 0.1 * csf * dxy[1])
    p[-2] = (max_xy[0] + 0.8 * csf * dxy[0], min_xy[1] - 0.1 * csf * dxy[1])
    p[-1] = (min_xy[0] + 0.5 * csf * dxy[0], max_xy[1] + 0.8 * csf * dxy[1])

    # triangle vertex indices
    t = np.empty((2 * n_p + 2, 3), dtype = int)  # empty space for triangle vertex indices
    t[0] = (n_p, n_p + 1, n_p + 2)  # initial set of one containing triangle
    nt = 1  # number of triangles so far populated

    # edges: list of indices of triangles and edge within triangle; -1 indicates no triangle using edge
    e = np.full((3 * n_p + 6, 2, 2), fill_value = -1, dtype = int)
    e[:3, 0, 0] = 0
    for edge in range(3):
        e[edge, 0, 1] = edge
    ne = 3  # number of edges so far in use

    # edge indices (in e) for each triangle, first axis indexed in sync with t
    te = np.empty((2 * n_p + 2, 3), dtype = int)  # empty space for triangle edge indices
    te[0] = (0, 1, 2)

    # mask tracking which edges have been flipped
    fm = np.zeros((3 * n_p + 6), dtype = bool)

    # debug plot here
    #   if plot_fn: plot_fn(p, t[:nt])

    progress_period = max(n_p // 100, 1)
    progress_count = progress_period

    for p_i in range(n_p):  # index of next point to consider

        if progress_fn is not None and progress_count <= 0:
            progress_fn(float(p_i) / float(n_p))
            progress_count = progress_period

        # find triangle that contains this point
        f_t = None

        for ti in range(nt):
            if vec.in_triangle_edged(p[t[ti, 0]], p[t[ti, 1]], p[t[ti, 2]], p[p_i]):
                f_t = ti
                break
        assert f_t is not None, 'failed to find triangle containing point'

        # take copy of edge indices for containing triangle
        e0, e1, e2 = te[f_t]

        # split containing triangle into 3, using new point as common vertex
        t[nt] = (t[f_t, 1], t[f_t, 2], p_i)
        t[nt + 1] = (t[f_t, 2], t[f_t, 0], p_i)
        t[f_t, 2] = p_i
        # add 3 new edges and update te
        e[ne, 0] = (f_t, 2)
        e[ne, 1] = (nt + 1, 1)
        e[ne + 1, 0] = (f_t, 1)
        e[ne + 1, 1] = (nt, 2)
        e[ne + 2, 0] = (nt, 1)
        e[ne + 2, 1] = (nt + 1, 2)
        if e[e1, 0, 0] == f_t:
            e[e1, 0] = (nt, 0)
        elif e[e1, 1, 0] == f_t:
            e[e1, 1] = (nt, 0)
        else:
            raise Exception('edge breakdown')
        if e[e2, 0, 0] == f_t:
            e[e2, 0] = (nt + 1, 0)
        elif e[e2, 1, 0] == f_t:
            e[e2, 1] = (nt + 1, 0)
        else:
            raise Exception('edge breakdown')
        te[nt] = (e1, ne + 2, ne + 1)
        te[nt + 1] = (e2, ne, ne + 2)
        te[f_t, 1] = ne + 1
        te[f_t, 2] = ne

        nt += 2
        ne += 3

        # now recursively try flipping sides
        fm[:] = False

        # debug plot here, with new point, before flipping
        if plot_fn is not None:
            plot_fn(p, t[:nt])

        flip(e0)
        flip(e1)
        flip(e2)

        progress_count -= 1

    # remove any triangles using invented container vertices
    tri_set = t[np.where(np.all(t[:nt] < n_p, axis = 1))]
    if plot_fn is not None:
        plot_fn(p, tri_set)

    external_t = t[np.where(np.any(t[:nt] >= n_p, axis = 1))]
    external_pi = np.unique(external_t)[:-3]  # discard 3 invented container vertices

    if progress_fn is not None:
        progress_fn(1.0)

    return tri_set, external_pi


def dt(p, algorithm = None, plot_fn = None, progress_fn = None, container_size_factor = 100.0, return_hull = False):
    """Returns the Delauney Triangulation of 2D point set p.

    arguments:
       p (numpy float array of shape (N, 2): the x,y coordinates of the points
       algorithm (string, optional): selects which algorithm to use; current options: ['simple'];
          if None, the current best algorithm is selected
       plot_fn (function of form f(p, t), optional): if present, this function is called each time the
          algorithm feels it is worth refreshing a plot of the progress; p is a copy of the point set,
          depending on the algorithm with 3 extra points added to form an enveloping triangle
       progress_fn (function of form f(x), optional): if present, this function is called at regulat
          intervals by the algorithm, passing increasing values in the range 0.0 to 1.0 as x
       container_size_factor (float, default 100.0): the larger this number, the more likely the
          resulting triangulation is to be convex; reduce to 1.0 to allow slight concavities
       return_hull (boolean, default False): if True, a pair is returned with the second item being
          a clockwise ordered list of indices into p identifying the points on the boundary of the
          returned triangulation

    returns:
       numpy int array of shape (M, 3) - being the indices into the first axis of p of the 3 points
          per triangle in the Delauney Triangulation - and if return_hull is True, another int array
          of shape (B,) - being indices into p of the clockwise ordered points on the boundary of
          the triangulation
    """
    assert p.ndim == 2 and p.shape[1] >= 2, 'bad points shape for 2D Delauney Triangulation'

    if not algorithm:
        algorithm = 'simple'

    if algorithm == 'simple':
        t, boundary = _dt_simple(p,
                                 plot_fn = plot_fn,
                                 progress_fn = progress_fn,
                                 container_size_factor = container_size_factor)
        if return_hull:
            return t, vec.clockwise_sorted_indices(p, boundary)
        else:
            return t
    else:
        raise Exception('unrecognised Delauney Triangulation algorithm name')


def ccc(p1, p2, p3):
    """Returns the centre of the circumcircle of the three points in the xy plane."""

    # two edges as vectors
    v12 = p2 - p1
    v13 = p3 - p1
    # midpoints of the two edges
    m12 = 0.5 * (p1 + p2)
    m13 = 0.5 * (p1 + p3)
    # pairs of points defining two edge normals passing through the midpoints
    o12 = m12.copy()
    o12[0] += v12[1]
    o12[1] -= v12[0]
    o13 = m13.copy()
    o13[0] += v13[1]
    o13[1] -= v13[0]
    return meet.line_line_intersect(m12[0], m12[1], o12[0], o12[1], m13[0], m13[1], o13[0], o13[1])


def voronoi(p: npt.NDArray[np.float_], t: npt.NDArray[np.int_], b: npt.NDArray[np.int_], aoi: rql.Polyline) -> \
        Tuple[npt.NDArray[np.float_], List[List[npt.NDArray[np.int_]]]]:
    """Returns dual Voronoi diagram for a Delauney triangulation.

    arguments:
       p (numpy float array of shape (N, 2)): seed points used in the Delauney triangulation
       t (numpy int array of shape (M, 3)): the Delauney triangulation of p as returned by dt()
       b (numpy int array of shape (B,)): clockwise sorted list of indices into p of the boundary
          points of the triangulation t
       aoi (lines.Polyline): area of interest; a closed clockwise polyline that must strictly contain
          all p (no points exactly on or outside the polyline)

    returns:
       c, v where: c is a numpy float array of shape (M+E, 2) being the circumcircle centres of
       the M triangles and E boundary points from the aoi polygon line; and v is a list of
       N Voronoi cell lists of clockwise ints, each int being an index into c

    notes:
       the aoi polyline forms the outer boundary for the Voronoi polygons for points on the
       outer edge of the triangulation; all points p must lie strictly within the aoi, which
       must be convex; the triangulation t, of points p, must also have a convex hull; note
       that the dt() function can produce a triangulation with slight concavities on the hull,
       especially for smaller values of its container_size_factor argument
    """

    # this code assumes that the Voronoi polygon for a seed point visits the circumcentres of
    # all the triangles that make use of the point – currently understood to be always the case
    # for a Delauney triangulation

    def __aoi_intervening_nodes(aoi_count, c_count, seg_a, seg_c):
        nodes = []
        seg = seg_a
        while seg != seg_c:
            seg = (seg + 1) % aoi_count
            nodes.append(c_count + seg)
        return nodes

    def __shorter_sides_p_i(p3):
        max_length = -1.0
        max_i = None
        for i in range(3):
            opp_length = vec.naive_length(p3[i - 1] - p3[i - 2])
            if opp_length > max_length:
                max_length = opp_length
                max_i = i
        return max_i

    def __azi_between(a, c, t):
        if c < a:
            c += 360.0
        return (a <= t <= c) or (a <= t + 360.0 <= c)

    def __seg_for_ci(ci):  # returns hull segment for a boundary index
        nonlocal ca_count, cah_count, caho_count, cahon_count, wing_hull_segments
        if ci < ca_count:
            return None
        if ci < cah_count:  # hull edge intersection
            return ci - ca_count
        if ci < caho_count:  # wings
            oi, wing = divmod(ci - cah_count, 2)
            return wing_hull_segments[oi, wing]
        if ci < cahon_count:  # virtual centre for hull edge
            return ci - caho_count
        # else virtual centre for hull point; arbitrarily pick clockwise segment
        return ci - cahon_count

    def __intervening_aoi_indices(aoi_count, aoi_intersect_segments, c_count, ca_count, cah_count, ci_for_p,
                                  out_pair_intersect_segments):
        # build list of intervening aoi boundary point indices and append to list
        aoi_nodes = []
        r = [0] if len(ci_for_p) == 2 else range(len(ci_for_p))
        just_done_pair = False
        for cii in r:
            cip = ci_for_p[cii]
            ci = ci_for_p[(cii + 1) % len(ci_for_p)]
            if cip >= c_count and ci >= c_count and not just_done_pair:
                # identify aoi segments
                if cip < cah_count:
                    aoi_seg_a = aoi_intersect_segments[cip - ca_count]
                else:
                    aoi_seg_a = out_pair_intersect_segments[divmod(cip - cah_count, 2)]
                if ci < cah_count:
                    aoi_seg_c = aoi_intersect_segments[ci - ca_count]
                else:
                    aoi_seg_c = out_pair_intersect_segments[divmod(ci - cah_count, 2)]
                aoi_nodes += __aoi_intervening_nodes(aoi_count, c_count, aoi_seg_a, aoi_seg_c)
                just_done_pair = True
            else:
                just_done_pair = False
        return aoi_nodes

    def __ci_non_hull(b_i, c, c_count, ci_for_p, p, p_i):
        trimmed_ci = []
        cii = 0
        finish_at = len(ci_for_p)
        while cii < finish_at:
            if ci_for_p[cii] < c_count:
                trimmed_ci.append(ci_for_p[cii])
                cii += 1
                continue
            start_cii = cii
            cii_seg = __seg_for_ci(ci_for_p[cii])
            while cii == 0 and ci_for_p[start_cii - 1] >= c_count and __seg_for_ci(ci_for_p[start_cii - 1]) == cii_seg:
                start_cii -= 1
            start_cii = start_cii % len(ci_for_p)
            end_cii = cii + 1
            while end_cii < len(ci_for_p) and ci_for_p[end_cii] >= c_count and __seg_for_ci(
                    ci_for_p[end_cii]) == cii_seg:
                end_cii += 1
            end_cii -= 1  # unpythonesque: end element included in scan
            if end_cii == start_cii:
                trimmed_ci.append(ci_for_p[cii])
                cii += 1
                continue
            if end_cii < start_cii:
                finish_at = start_cii
            if end_cii == (start_cii + 1) % len(ci_for_p):
                trimmed_ci.append(ci_for_p[start_cii])
                trimmed_ci.append(ci_for_p[end_cii])
                cii = end_cii + 1
                continue
            start_azi = vec.azimuth(c[ci_for_p[start_cii]] - p[p_i, :2])
            end_azi = vec.azimuth(c[ci_for_p[end_cii]] - p[p_i, :2])
            scan_cii = start_cii
            while True:
                if scan_cii == start_cii:
                    trimmed_ci.append(ci_for_p[scan_cii])
                elif scan_cii == end_cii:
                    trimmed_ci.append(ci_for_p[scan_cii])
                    break
                else:
                    # if point is around a hull corner from previous, then include (?)
                    next_cii = (scan_cii + 1) % len(ci_for_p)
                    if b_i is None and __seg_for_ci(ci_for_p[scan_cii]) != __seg_for_ci(ci_for_p[next_cii]):
                        trimmed_ci.append(ci_for_p[scan_cii])
                        trimmed_ci.append(ci_for_p[next_cii])
                        scan_cii = next_cii
                    elif b_i is None and __seg_for_ci(trimmed_ci[-1]) != __seg_for_ci(ci_for_p[scan_cii]):
                        trimmed_ci.append(ci_for_p[scan_cii])
                    else:
                        azi = vec.azimuth(c[ci_for_p[scan_cii]] - p[p_i, :2])
                        if __azi_between(start_azi, end_azi, azi):
                            trimmed_ci.append(ci_for_p[scan_cii])
                if scan_cii == end_cii:
                    break
                scan_cii = (scan_cii + 1) % len(ci_for_p)
            cii = end_cii + 1
        return trimmed_ci

    def __closest_to_seed(c, c_count, ci_for_p, hull_node_azi, p, p_i):
        best_a_i = None
        best_c_i = None
        best_a_azi = -181.0
        best_c_azi = 181.0
        for cii, val in enumerate(ci_for_p):
            if val < c_count:
                continue
            azi = vec.azimuth(c[val] - p[p_i, :2]) - hull_node_azi
            if azi > 180.0:
                azi -= 360.0
            elif azi < -180.0:
                azi += 360.0
            if 0.0 > azi > best_a_azi:
                best_a_azi = azi
                best_a_i = cii
            elif 0.0 <= azi < best_c_azi:
                best_c_azi = azi
                best_c_i = cii
        assert best_a_i is not None and best_c_i is not None
        trimmed_ci = []
        for cii, val in enumerate(ci_for_p):
            if val < c_count or cii == best_a_i or cii == best_c_i:
                trimmed_ci.append(val)
        return trimmed_ci

    def __ci_replace(c_count, ca_count, cah_count, caho_count, cahon_count, ci_for_p, p, p_i, t, tc_outwith_aoi):
        # where circumcirle (or virtual) centre is outwith aoi, replace with a point on aoi boundary
        # virtual centres related to hull points (not hull edges) can be discarded
        trimmed_ci = []
        for ci in ci_for_p:
            if ci < c_count:  # genuine triangle
                if ci in tc_outwith_aoi:  # replace with one or two wing normal intersection points
                    oi = tc_outwith_aoi.index(ci)
                    wing_i = cah_count + 2 * oi
                    shorter_t_i = __shorter_sides_p_i(p[t[ci]])
                    if t[ci, shorter_t_i] == p_i:
                        trimmed_ci += [wing_i, wing_i + 1]
                    elif t[ci, shorter_t_i - 1] == p_i:
                        trimmed_ci.append(wing_i)
                    else:
                        trimmed_ci.append(wing_i + 1)
                else:
                    trimmed_ci.append(ci)
            elif ci < cahon_count:
                # extended virtual centre for a hull edge (discard hull point virtual centres)
                # replace with index for intersection point on aoi boundary
                trimmed_ci.append(ca_count + ci - caho_count)
        return trimmed_ci

    def __veroni_cells(aoi_count, aoi_intersect_segments, b, c, c_count, ca_count, cah_count, caho_count, cahon_count,
                       hull_count, out_pair_intersect_segments, p, t, tc_outwith_aoi):
        # list of voronoi cells (each a numpy list of node indices into c extended with aoi points etc)
        v = []
        # for each seed point build the voronoi cell
        for p_i in range(len(p)):

            # find triangles making use of that point
            ci_for_p = list(np.where(t == p_i)[0])

            # if seed point is on hull boundary, introduce three extended virtual centres
            b_i = None
            if p_i in b:
                b_i = np.where(b == p_i)[0][0]  # index into hull coordinates
                p_b_i = (b_i - 1) % hull_count  # predecessor, ie. anti-clockwise boundary point
                ci_for_p += [caho_count + p_b_i, cahon_count + b_i, caho_count + b_i]

            # find azimuths of vectors from seed point to circumcircle centres (and virtual centres)
            azi = [vec.azimuth(centre - p[p_i, :2]) for centre in c[ci_for_p, :2]]
            # if this is a hull seed point, make a note of azimuth to virtual centre
            hull_node_azi = None if b_i is None else azi[-2]
            # sort triangle indices for seed point into clockwise order of circumcircle (and virtual) centres
            ci_for_p = [ti for (_, ti) in sorted(zip(azi, ci_for_p))]

            ci_for_p = __ci_replace(c_count, ca_count, cah_count, caho_count, cahon_count, ci_for_p, p, p_i, t,
                                    tc_outwith_aoi)

            # if this is a hull seed point, classify aoi boundary points into anti-clockwise or clockwise, and find
            # closest to seed
            if b_i is not None:
                ci_for_p = __closest_to_seed(c, c_count, ci_for_p, hull_node_azi, p, p_i)

            # for sequences on aoi boundary, just keep those between the first and last (?)

            #      elif any([ci < c_count for ci in ci_for_p]):
            else:
                ci_for_p = __ci_non_hull(b_i, c, c_count, ci_for_p, p, p_i)

            # reverse points if needed for pair of aoi points only
            assert len(ci_for_p) >= 2
            if len(ci_for_p) == 2:
                seg_0 = __seg_for_ci(ci_for_p[0])
                seg_1 = __seg_for_ci(ci_for_p[1])
                if seg_0 is not None and seg_1 is not None and seg_0 == (seg_1 + 1) % hull_count:
                    ci_for_p.reverse()

            ci_for_p += __intervening_aoi_indices(aoi_count, aoi_intersect_segments, c_count, ca_count, cah_count,
                                                  ci_for_p, out_pair_intersect_segments)

            #  remove circumcircle centres that are outwith area of interest
            ci_for_p = np.array([ti for ti in ci_for_p if ti >= c_count or ti not in tc_outwith_aoi], dtype = int)

            # find azimuths of vectors from seed point to circumcircle centres and aoi boundary points
            azi = [vec.azimuth(centre - p[p_i, :2]) for centre in c[ci_for_p, :2]]

            # re-sort triangle indices for seed point into clockwise order of circumcircle centres and boundary points
            ordered_ci = [ti for (_, ti) in sorted(zip(azi, ci_for_p))]

            v.append(ordered_ci)
        return v

    # log.debug(f'\n\nVoronoi: nt: {len(p)}; nt: {len(t)}; hull: {len(b)}; aoi: {len(aoi.coordinates)}')
    # todo: allow aoi to be None in which case create an aoi as hull with border

    assert p.ndim == 2 and p.shape[0] > 2 and p.shape[1] >= 2
    assert t.ndim == 2 and t.shape[1] == 3
    assert b.ndim == 1 and b.shape[0] > 2
    assert len(aoi.coordinates) >= 3
    assert aoi.isclosed
    assert aoi.is_clockwise()

    # create temporary polyline for hull of triangulation
    hull = rql.Polyline(
        aoi.model,
        set_bool = True,  # polyline is closed
        set_coord = p[b],
        set_crs = aoi.crs_uuid,
        title = 'triangulation hull')
    hull_count = len(b)

    # check for concavities in hull
    if not hull.is_convex():
        log.warning('Delauney triangulation is not convex; Voronoi diagram construction might fail')

    # compute circumcircle centres
    c = np.zeros((t.shape[0], 2))
    for ti in range(len(t)):
        c[ti] = ccc(p[t[ti, 0]], p[t[ti, 1]], p[t[ti, 2]])
    c_count = len(c)

    # make list of triangle indices whose circumcircle centres are outwith the area of interest
    tc_outwith_aoi = [ti for ti in range(c_count) if not aoi.point_is_inside_xy(c[ti])]
    o_count = len(tc_outwith_aoi)

    # make space for combined points data needed for all voronoi cell nodes:
    # 1. circumcircle centres for triangles in delauney triangulation
    # 2. nodes defining area of interest polygon
    # 3. intersection of normals to triangulation hull edges with aoi polygon
    # 4. extra intersections for normals to other two (non-hull) triangle edges, with aoi,
    #    where circumcircle centre is outside the area of interest
    # 5. extended ccc for hull edge normals
    # 6. extended ccc for hull points
    # (5 & 6 only used during construction)
    c = np.concatenate((c, aoi.coordinates[:, :2], np.zeros(
        (hull_count, 2), dtype = float), np.zeros(
            (2 * o_count, 2), dtype = float), np.zeros((hull_count, 2),
                                                       dtype = float), np.zeros((hull_count, 2), dtype = float)))
    aoi_count = len(aoi.coordinates)
    ca_count = c_count + aoi_count
    cah_count = ca_count + hull_count
    caho_count = cah_count + 2 * o_count
    cahon_count = caho_count + hull_count
    assert cahon_count + hull_count == len(c)

    #  compute intersection points between hull edge normals and aoi polyline
    # also extended virtual centres for hull edges
    extension_scaling = 1000.0 * np.sum((np.max(aoi.coordinates, axis = 0) - np.min(aoi.coordinates, axis = 0))[:2])
    aoi_intersect_segments = np.empty((hull_count,), dtype = int)
    for ei in range(hull_count):
        # use segment midpoint and normal methods of hull to project out
        m = hull.segment_midpoint(ei)[:2]  # midpoint
        norm_vec = hull.segment_normal(ei)[:2]
        n = m + norm_vec  # point on normal
        # use first intersection method of aoi to intersect projected normal from triangulation hull
        aoi_seg, aoi_x, aoi_y = aoi.first_line_intersection(m[0], m[1], n[0], n[1], half_segment = True)
        assert aoi_seg is not None
        # inject intersection points to extension area of c and take note of aoi segment of intersection
        c[ca_count + ei] = (aoi_x, aoi_y)
        aoi_intersect_segments[ei] = aoi_seg
        # inject extended virtual circle centres for hull edges, a long way out
        c[caho_count + ei] = c[ca_count + ei] + extension_scaling * norm_vec

    # compute extended virtual centres for hull nodes
    for ei in range(hull_count):
        pei = (ei - 1) % hull_count
        vector = vec.unit_vector(hull.segment_normal(pei)[:2] + hull.segment_normal(ei)[:2])
        c[cahon_count + ei] = hull.coordinates[ei, :2] + extension_scaling * vector

    # where cicrumcircle centres are outwith aoi, compute intersections of normals of wing edges with aoi
    out_pair_intersect_segments = np.empty((o_count, 2), dtype = int)
    wing_hull_segments = np.empty((o_count, 2), dtype = int)
    for oi, ti in enumerate(tc_outwith_aoi):
        tpi = __shorter_sides_p_i(p[t[ti]])
        for wing in range(2):
            # note: triangle nodes are anticlockwise
            m = 0.5 * (p[t[ti, tpi - 1]] + p[t[ti, tpi]])[:2]  # triangle edge midpoint
            edge_v = p[t[ti, tpi]] - p[t[ti, tpi - 1]]
            n = m + np.array((-edge_v[1], edge_v[0]))  # point on perpendicular bisector of triangle edge
            o_seg, o_x, o_y = aoi.first_line_intersection(m[0], m[1], n[0], n[1], half_segment = True)
            c[cah_count + 2 * oi + wing] = (o_x, o_y)
            out_pair_intersect_segments[oi, wing] = o_seg
            wing_hull_segments[oi, wing], _, _ = hull.first_line_intersection(m[0],
                                                                              m[1],
                                                                              n[0],
                                                                              n[1],
                                                                              half_segment = True)
            tpi = (tpi + 1) % 3

    v = __veroni_cells(aoi_count, aoi_intersect_segments, b, c, c_count, ca_count, cah_count, caho_count, cahon_count,
                       hull_count, out_pair_intersect_segments, p, t, tc_outwith_aoi)

    return c[:caho_count], v


def triangulated_polygons(p, v, centres = None):
    """Returns triangulation of polygons using centres as extra points.

    arguments:
       p (2D numpy float array): points used as vertices of polygons
       v (list of list of ints): ordered indices into p for each polygon
       centres (2D numpy float array, optional): the points to use as the centre for each polygon

    returns:
       points, triangles where: points is a copy of p extended with the centre points of polygons;
       and triangles is a numpy int array of shape (N, 3) being the triangulation of points, where N is
       equal to the overall length of v

    notes:
       if no centres are provided, balanced centre points are computed for the polygons;
       the polygons must be convex (at least from the perspective of the centre points);
       the clockwise/anti-clockwise order of the triangle edges will match that of the polygon;
       the centre point is the first point in each triangle;
       the order of triangles will match the order of vertices in a flattened view of list v;
       p and centres may have a shape of 2 or 3 in the second dimension (xy or xyz data);
       p & v could be the values (c, v) returned by the voronoi() function, in which case the
       original seed points p passed into voronoi() can be passed as centres here
    """

    assert p.ndim == 2 and p.shape[1] in [2, 3]
    assert len(v) > 0
    if centres is not None:
        assert centres.ndim == 2
        assert len(centres) == len(v)
        assert centres.shape[1] == p.shape[1]

    model = rq.Model(create_basics = True)  # temporary object for handling Polylines
    crs = rqc.Crs(model)

    points = np.zeros((len(p) + len(v), p.shape[1]))  # polygon nodes, to be extended with centre points
    points[:len(p)] = p
    if centres is not None:
        points[len(p):] = centres

    t_count = sum([len(x) for x in v])
    triangles = np.empty((t_count, 3), dtype = int)
    t_index = 0

    for cell, poly_vertices in enumerate(v):
        # add polygon centre points to points array, if not already provided
        centre_i = len(p) + cell
        if centres is None:
            polygon = rql.Polyline(model,
                                   set_coord = p[np.array(poly_vertices, dtype = int)],
                                   set_bool = True,
                                   set_crs = crs.uuid,
                                   title = 'v cell')
            poly_centre = polygon.balanced_centre()
            points[centre_i] = poly_centre[:p.shape[1]]
        # and populate triangles for this polygon
        for ti in range(len(poly_vertices)):
            triangles[t_index] = (centre_i, poly_vertices[ti], poly_vertices[(ti + 1) % len(poly_vertices)])
            t_index += 1
    assert t_index == t_count

    return points, triangles



import pytest
from packaging.version import Version

import resqpy
from resqpy import model


def test_empty_model():
    _ = model.Model()
    return


def test_all_imports():
    from resqpy import (crs, derived_model, fault, grid, grid_surface, lines, organize, property, rq_import, surface,
                        time_series, well)

    # The line below prevents IDEs from deleting the above
    _ = (crs, derived_model, fault, grid, grid_surface, lines, organize, property, rq_import, surface, time_series,
         well)

    #    from resqpy.olio import *
    return


def test_version():
    # This is dynamically created when package is built
    # If this fails when running tests locally, ensure you have installed the dev dependencies specified in setup.cfg
    # In particular, try:  pip install setuptools_scm
    version_string = resqpy.__version__

    # Ensure version string is a PEP-440 compliant version
    version = Version(version_string)
    assert version >= Version("1.1.1")



import os
from inspect import getsourcefile
import numpy as np

import resqpy.model as rq


def write_snapshot(data, title):
    # Method to write out a new snapshot of data if one is needed
    reshaped_array = data.reshape(data.shape[0], -1)
    np.savetxt(f'C:\\Test\\{title}.txt', reshaped_array)


def check_load_snapshot(data, filename):
    # Compare the actual data against the stored expected array
    loaded_array = np.loadtxt(filename)
    expected_array = loaded_array.reshape(loaded_array.shape[0], loaded_array.shape[1] // data.shape[2], data.shape[2])
    np.testing.assert_array_almost_equal(data, expected_array)


def test_check_transmisibility_output(test_data_path):
    current_filename = os.path.split(getsourcefile(lambda: 0))[0]
    base_folder = os.path.dirname(os.path.dirname(current_filename))

    resqml_file_root = base_folder + '/example_data/block.epc'
    grid_model = rq.Model(resqml_file_root)
    resqml_grid = grid_model.grid()
    k, j, i = resqml_grid.transmissibility()

    snapshot_filename = current_filename + "/snapshots/transmissibility/"
    check_load_snapshot(i, f'{snapshot_filename}block_i.txt')
    check_load_snapshot(j, f'{snapshot_filename}block_j.txt')
    check_load_snapshot(k, f'{snapshot_filename}block_k.txt')
